#-------------------------------------------------------------------------------

# Compute expression based on a single similarity slider.
# Try to do simple optimizations based on scale and exponent values.
#
proc grSimOne {v f n t} {
  global parm
  if { $parm($n,s) == 0 } { return 0 }
  switch $t {
    1 {
       set cmd ""
       set diff "(1.0 - abs(na($f,$n) - na($v,$n)))"
       if { $parm($n,s) != 1 } {
          append cmd "$parm($n,s) * "
       }
       if { $parm($n,e) != 1 } {
          append cmd "pow( $diff, $parm($n,e))"
       } else {
          append cmd $diff
       }
    }
    0 {
       set cmd "a($f,$n) == a($v,$n) ? $parm($n,s) : 0"
    }
  }
  return $cmd
}

#-------------------------------------------------------------------------------

# Compute sum based on all sliders.
# Again, optimize in terms of exponent value.
#
proc grSimExpr {v f} {
   global parm
   set sum ""
   foreach s $parm(sliders) {
      set n [lindex $s 0]
      set t [lindex $s 1]

      append sum "( " [grSimOne $v $f $n $t] " ) + "
   }

   if { $parm(e) != 1 } {
      set cmd "pow( $sum 0, $parm(e) ) * mag($f)"
   } else {
      set cmd "($sum 0 ) * mag($f)"
   }

   return $cmd
}

#-------------------------------------------------------------------------------

# Create slider or slider pair for similarity dialog box.
#
proc grSimSlider {w n numeric label} {
   global parm
   frame $w.f$n
   scale $w.f$n.s -label "$label scale" -from -1.0 -to 1.0 \
      -length 2c -orient horizontal -resolution 0.01 \
      -font -Adobe-Helvetica-Medium-R-Normal--*-80-*-*-*-*-*-*
   $w.f$n.s set $parm($n,s)
   $w.f$n.s configure -command "SetSlider $n s"
   if { $numeric } {
      scale $w.f$n.e -label exponent -from 0.0 -to 10.0 \
         -length 2c -orient horizontal -resolution 0.1 \
         -font -Adobe-Helvetica-Medium-R-Normal--*-80-*-*-*-*-*-*
      $w.f$n.e set $parm($n,e)
      $w.f$n.e configure -command "SetSlider $n e"
      pack $w.f$n.s $w.f$n.e -side left
   } else {
      pack $w.f$n.s -side left
   }
}

#-------------------------------------------------------------------------------

#  Build complete Similarity dialog box.
#
proc grSimDlg {} {
   global parm

   frame .t.fe
   scale .t.fe.e -label Exponent -from 0.0 -to 10.0 \
      -length 2c -orient horizontal -resolution 0.1 \
      -font -Adobe-Helvetica-Medium-R-Normal--*-80-*-*-*-*-*-* \
      -command "SetSlider {} e"
   .t.fe.e set $parm(e)
   pack .t.fe.e -side left
   pack .t.fe -anchor w

   foreach s $parm(sliders) {
      set n [lindex $s 0]
      set t [lindex $s 1]
      set l [lindex $s 2]
      grSimSlider .t $n $t $l
      pack .t.f$n -side top -anchor w
   }

   frame .t.fop
   button .t.fop.none  -text None  -command {SetSliders 0} \
      -font -Adobe-Helvetica-Medium-R-Normal--*-100-*-*-*-*-*-*
   button .t.fop.all   -text All   -command {SetSliders 1} \
      -font -Adobe-Helvetica-Medium-R-Normal--*-100-*-*-*-*-*-*
   button .t.fop.reset -text Reset -command {SetSliders 2} \
      -font -Adobe-Helvetica-Medium-R-Normal--*-100-*-*-*-*-*-*
   pack .t.fop.all .t.fop.none .t.fop.reset -side left
   pack .t.fop -side top
}

proc SetSliders {mode} {
   global parm
   .t.fe.e set 1
   foreach s $parm(sliders) {
      set n [lindex $s 0]
      set t [lindex $s 1]
      switch $mode {
         0 { .t.f$n.s set 0 }
         1 { .t.f$n.s set 1 }
         2 { .t.f$n.s set $parm($n,s) }
      }
      if { $t == 1 } {
         switch $mode {
            0 { .t.f$n.e set 1 }
            1 { .t.f$n.e set 1 }
            2 { .t.f$n.e set $parm($n,e) }
         }
      }
   }
}

#-------------------------------------------------------------------------------

# Set colour of slider text and trough based on value set.
#
proc SetSlider {n suffix value} {
   global parm
   if { $n == "" } {
      set w .t.fe.e;		set k "e"
   } else {
      set w .t.f$n.$suffix;	set k "$n,$suffix"
   }

   if { $suffix == "s" } {
      if { $value > 0 } {
         set clr [HSVtoRGB 120 [expr 0.2 + 0.8 * pow($value,3)] 1]
      } else {
         if { $value < 0 } {
            set clr [HSVtoRGB 0 [expr 0.2 + 0.8 * pow($value * -1.0,3)] 1]
         } else {
            set clr gray
         }
      }
      $w configure -troughcolor $clr
      #if { $value == 1 || $value == 0 } {
      #   $w configure -troughcolor gray
      #} else {
      #   $w configure -troughcolor lightblue
      #}
      if { $value != $parm($k) } {
         $w configure -fg blue
      } else {
         $w configure -fg black
      }
   } else {
      #if { $value == 1 } {
      #   $w configure -troughcolor gray
      #} else {
      #   $w configure -troughcolor lightblue
      #}
      if { $value > 1 } {
         set clr [HSVtoRGB 120 [expr 0.2 + 0.8 * pow($value / 10.0,3)] 1]
      } else {
         if { $value < 1 } {
            set clr [HSVtoRGB 225 [expr 1.0 - (0.2 + 0.8 * pow($value,3))] 1]
         } else {
            set clr gray
         }
      }
      $w configure -troughcolor $clr
      if { $parm($k) != $value } {
         $w configure -fg blue
      } else {
         $w configure -fg black
      }
   }
}

#-------------------------------------------------------------------------------

# Copy slider settings and recompute view.
#
proc grSimCmd {} {
   global parm
   set parm(e) [.t.fe.e get]
   foreach s $parm(sliders) {
      set n [lindex $s 0]
      set t [lindex $s 1]
      set parm($n,s) [.t.f$n.s get]
      if { $t == 1 } { set parm($n,e) [.t.f$n.e get] }
   }
   RecomputeView 1 0 0
}

#-------------------------------------------------------------------------------

# Initialize slider values.
#
proc grInitSliders {} {
   global parm
   set parm(e) 1
   foreach s $parm(sliders) {
      set n [lindex $s 0]
      set t [lindex $s 1]
      set l [lindex $s 2]
      set parm($n,s) 1
      set parm($n,e) 1	;# must create for grSimOne even though don't use...
   }
}

#-------------------------------------------------------------------------------
