#include <stdio.h>
#include <assert.h>
#include <tk.h>

#include <hygraph/Hygraph.h>
#include <util/MN_Series.h>
#include <util/MN_Dictionary.h>
#include <util/MN_Set.h>
#include <util/MN_Color.h>
#include <util/unix.h>
#include <server/server.h>
#include <server/hgInfo.h>

#include "misc.h"
#include "util.h"
#include "tclTkHInfo.h"

extern TclTkHInfo* current_hi;

//------------------------------------------------------------------------------

// Func: d(Index)	distance in matrix Index
// Func: nd(Index)	normalized distance in matrix Index
//
/*
MATH_FUNC( grDist )
{
   CHK_CUR;
   result->type = TCL_DOUBLE;
   if( fpcursor == NIL ) RET_ERROR("bad context")
   switch( (int)data )
   {
      case 1:
         result->doubleValue = current_hi->Dist( args[0].intValue, true  );
         break;
      case 0:
      default:
         result->doubleValue = current_hi->Dist( args[0].intValue, false );
         break;
   }
   return TCL_OK;
}
*/

// Func: fm($f)		focal point magnification
//
MATH_FUNC( grMag )
{
   CHK_CUR;
   result->type = TCL_DOUBLE;
   Vertex* v = (Vertex*)args[0].intValue;
   if( DQ(type,v) != 0 ) RET_ERROR("object not vertex")
   result->doubleValue = DQ(mag,v);
   return TCL_OK;
}

// Func: sp($u,$v,$p)		shortest path length
//
MATH_FUNC( grSP )
{
   CHK_CUR;
   result->type = TCL_DOUBLE;

   Vertex* u = (Vertex*)args[0].intValue;
   if( DQ(type,u) != 0 ) RET_ERROR("bad variable 1")

   Vertex* v = (Vertex*)args[1].intValue;
   if( DQ(type,v) != 0 ) RET_ERROR("bad variable 2")

   result->doubleValue =
      current_hi->MinWtPaths( (int)args[2].intValue, u, v );
   return TCL_OK;
}

// Func: onP($o,$p)
//
MATH_FUNC( grOnP )
{
   CHK_CUR;
   result->type = TCL_INT;
   HGObject* o = (HGObject*)args[0].intValue;
   int t = DQ(type,o);
   result->intValue = current_hi->OnPath( o, args[1].intValue, t );
   return TCL_OK;
}

//------------------------------------------------------------------------------

// Func: gw()		graph width
// Func: gh()		graph height
MATH_FUNC( grGetGeom2 )
{
   CHK_CUR;
   result->type = TCL_DOUBLE;

   HGContainer* c = current_hi->hygraph->c;
   result->doubleValue = (int)data ? c->size.y : c->size.x;
   return TCL_OK;
}

// Func: x($o)		horizontal position
// Func: y($o)		vertical position
// Func: w($o)		width
// Func: h($o)		height
//
MATH_FUNC( grGetGeom )
{
   CHK_CUR;
   result->type = TCL_DOUBLE;

   HGObject* o = (HGObject*)args[0].intValue;
   if( DQ(type,o) != 0 ) RET_ERROR("object not vertex")

   int type = (int)data;
   float v = 0;
   HGContainer* c = o->c;
   switch( type )
   {
      case 0: v = c->pos.x;  break;
      case 1: v = c->pos.y;  break;
      case 2: v = c->size.x; break;
      case 3: v = c->size.y; break;
      default: RET_ERROR("internal error")
   }
   result->doubleValue = v;
   return TCL_OK;
}

//------------------------------------------------------------------------------

int GrMath_Init( Tcl_Interp* interp )
{
   Tcl_ValueType a_i[] = { TCL_INT };
   Tcl_ValueType a_ii[] = { TCL_INT, TCL_INT };
   Tcl_ValueType a_iii[] = { TCL_INT, TCL_INT, TCL_INT };

   Tcl_CreateMathFunc( interp, "sp",  3, a_iii, grSP,     (ClientData)NULL );
   Tcl_CreateMathFunc( interp, "onP", 2, a_ii,  grOnP,    (ClientData)0 );

   Tcl_CreateMathFunc( interp, "x",   1, a_i, grGetGeom,  (ClientData)0 );
   Tcl_CreateMathFunc( interp, "y",   1, a_i, grGetGeom,  (ClientData)1 );
   Tcl_CreateMathFunc( interp, "w",   1, a_i, grGetGeom,  (ClientData)2 );
   Tcl_CreateMathFunc( interp, "h",   1, a_i, grGetGeom,  (ClientData)3 );

   Tcl_CreateMathFunc( interp, "w",   0, a_i, grGetGeom2, (ClientData)0 );
   Tcl_CreateMathFunc( interp, "h",   0, a_i, grGetGeom2, (ClientData)1 );

   Tcl_CreateMathFunc( interp, "mag", 1, a_i, grMag,      (ClientData)NULL );

/*
   Tcl_CreateMathFunc( interp, "d",   1, a_i, grDist,     (ClientData)0 );
   Tcl_CreateMathFunc( interp, "nd",  1, a_i, grDist,     (ClientData)1 );
*/

   return TCL_OK;
}

//------------------------------------------------------------------------------
