#include <stdio.h>
#include <assert.h>
#include <tk.h>

#include <hygraph/Hygraph.h>
#include <util/MN_Series.h>
#include <util/MN_Dictionary.h>
#include <util/MN_Set.h>
#include <util/MN_Color.h>
#include <util/unix.h>
#include <server/server.h>
#include <server/hgInfo.h>

#include "misc.h"
#include "util.h"
#include "tclTkHInfo.h"

extern TclTkHInfo* current_hi;

#define PFMT(P) "0x%lx",(long unsigned int)P

//------------------------------------------------------------------------------

COMMAND( grField )
{
   char buf[8];
   HGObject* o;

   CHK_CUR;
   CHK_ARGS( !=, 2, "usage: grField var" );

   switch( (int)data )
   {
   case 'I':
      o = (HGObject*)strtol( argv[1], (char**)NULL, 16 );
      Tcl_AppendResult( interp, current_hi->hygraph->Id(o), 0 );
      break;
   case 'L':
      o = (HGObject*)strtol( argv[1], (char**)NULL, 16 );
      Tcl_AppendResult( interp, current_hi->hygraph->Label(o), 0 );
      break;
   case 'F':
      o = (HGObject*)strtol( argv[1], (char**)NULL, 16 );
      if( DQ(type,o) == 1 )
         sprintf( buf, PFMT(current_hi->hygraph->From((AArc*)o)) );
      else RET_ERROR("object not arc");
      Tcl_AppendResult( interp, buf, 0 );
      break;
   case 'T':
      o = (HGObject*)strtol( argv[1], (char**)NULL, 16 );
      if( DQ(type,o) == 1 )
         sprintf( buf, PFMT(current_hi->hygraph->To((AArc*)o)) );
      else RET_ERROR("object not arc");
      Tcl_AppendResult( interp, buf, 0 );
      break;
   case 'B':
      o = (HGObject*)strtol( argv[1], (char**)NULL, 16 );
      if( DQ(type,o) == 0 )
         sprintf( buf, PFMT(current_hi->hygraph->B((Vertex*)o)) );
      else RET_ERROR("object not vertex");
      Tcl_AppendResult( interp, buf, 0 );
      break;
   case 'V':
      o = (HGObject*)strtol( argv[1], (char**)NULL, 16 );
      if( DQ(type,o) == 2 )
         sprintf( buf, PFMT(current_hi->hygraph->From((Blob*)o)) );
      else RET_ERROR("object not blob");
      Tcl_AppendResult( interp, buf, 0 );
      break;
   default:
      RET_ERROR("internal error")
      break;
   }
   return TCL_OK;
}

//------------------------------------------------------------------------------

COMMAND( grSubRel )
{
   char buf[16];
   int i;
   Vertex* v;
   AArc* a;
   Blob* b;

   CHK_CUR;
   CHK_ARGS( !=, 2, "usage: grSubRel var" );

   switch( (int)data )
   {
   case 'I':
      v = (Vertex*)strtol( argv[1], (char**)NULL, 16 );
      if( DQ(type,v) != 0 ) RET_ERROR("object not vertex")
      for( i = 0; (a = current_hi->hygraph->In(v,i)) != NIL; i++ )
      {
         sprintf( buf, PFMT(a) );
         Tcl_AppendElement( interp, buf );
      }
      break; 
   case 'O':
      v = (Vertex*)strtol( argv[1], (char**)NULL, 16 );
      if( DQ(type,v) != 0 ) RET_ERROR("object not vertex")
      for( i = 0; (a = current_hi->hygraph->Out(v,i)) != NIL; i++ )
      {
         sprintf( buf, PFMT(a) );
         Tcl_AppendElement( interp, buf );
      }
      break;
   case 'B':
      v = (Vertex*)strtol( argv[1], (char**)NULL, 16 );
      if( DQ(type,v) != 0 ) RET_ERROR("object not vertex")
      for( i = 0; (b = current_hi->hygraph->B(v,i)) != NIL; i++ )
      {
         sprintf( buf, PFMT(b) );
         Tcl_AppendElement( interp, buf );
      }
      break;
   case 'V':
      b = (Blob*)strtol( argv[1], (char**)NULL, 16 );
      if( DQ(type,b) != 2 ) RET_ERROR("object not blob")
      for( i = 0; (v = current_hi->hygraph->To(b,i)) != NIL; i++ )
      {
         sprintf( buf, PFMT(v) );
         Tcl_AppendElement( interp, buf );
      }
      break;
   case 'A':
      v = (Vertex*)strtol( argv[1], (char**)NULL, 16 );
      if( DQ(type,v) != 0 ) RET_ERROR("object not vertex")
      while( (b = current_hi->hygraph->B(v)) != NIL )
      {
         v = current_hi->hygraph->From( b );
         sprintf( buf, PFMT(v) );
         Tcl_AppendElement( interp, buf );
      }
      break;
   default:
      RET_ERROR("internal error")
      break;
   }
   return TCL_OK;
}

//------------------------------------------------------------------------------

COMMAND( grRel )
{
   char buf[16];
   int i;
   Vertex* v;
   AArc* a;
   Blob* b;

   CHK_CUR;
   CHK_ARGS( !=, 1, "usage: grRel [V|A|B|F|S|E]" );

   switch( (int)data )
   {
   case 'V':
      for( i = 0; (v = current_hi->hygraph->V(i)) != NIL; i++ )
      {
         sprintf( buf, PFMT(v) );
         Tcl_AppendElement( interp, buf );
      }
      break; 
   case 'A':
      for( i = 0; (a = current_hi->hygraph->A(i)) != NIL; i++ )
      {
         sprintf( buf, PFMT(a) );
         Tcl_AppendElement( interp, buf );
      }
      break;
   case 'B':
      for( i = 0; (b = current_hi->hygraph->B(i)) != NIL; i++ )
      {
         sprintf( buf, PFMT(b) );
         Tcl_AppendElement( interp, buf );
      }
      break;
   case 'F':
      for( i = 0; (v = current_hi->fpts[i]) != NIL; i++ )
      {
         sprintf( buf, PFMT(v) );
         Tcl_AppendElement( interp, buf );
      }
      break;
   case 'S':
      for( i = 0; (v = current_hi->sources[i]) != NIL; i++ )
      {
         sprintf( buf, PFMT(v) );
         Tcl_AppendElement( interp, buf );
      }
      break;
   case 'E':
      for( i = 0; (v = current_hi->sinks[i]) != NIL; i++ )
      {
         sprintf( buf, PFMT(v) );
         Tcl_AppendElement( interp, buf );
      }
      break;
   default:
      RET_ERROR("internal error")
      break;
   }
   return TCL_OK;
}

//------------------------------------------------------------------------------

COMMAND( grSet )
{
   CHK_CUR;
   CHK_ARGS( !=, 4, "usage: grSet var index value" )

   HGObject* o = (HGObject*)strtol( argv[1], (char**)NULL, 16 );
   int i, ty = DQ(type,o);
   if( sscanf( argv[2], "%d", &i ) != 1 ) RET_ERROR("bad index")
   DQ(val[i],o) = atof( argv[3] );
   current_hi->sdirty[ty][i] = 1;
   return TCL_OK;
}
//------------------------------------------------------------------------------

COMMAND( grSelect )
{
   int listArgc, i, result;
   char **listArgv;

   CHK_CUR;
   CHK_ARGS( !=, 4, "usage: grSelect var list expr" )

   // Break the list up into elements, and execute the command once
   // for each value of the element.
   //

   result = Tcl_SplitList(interp, argv[2], &listArgc, &listArgv);
   if( result != TCL_OK ) return result;

   int* match = new int[listArgc];

   for( i = 0; i < listArgc; i++ )
   {
      if (Tcl_SetVar(interp, argv[1], listArgv[i], 0) == NULL)
      {
         Tcl_SetResult(interp, "couldn't set loop variable", TCL_STATIC);
         result = TCL_ERROR;
         break;
      }

      result = Tcl_ExprBoolean( interp, argv[3], &match[i] );
      if( result != TCL_OK ) break;
   }
   for( i = 0; i < listArgc; i++ )
      if( match[i] )
         Tcl_AppendElement( interp, listArgv[i] );
   delete match;
   ckfree((char *) listArgv);
   return result;
}

//------------------------------------------------------------------------------

COMMAND( grAgOp )
{
   int listArgc, i, result;
   char **listArgv;
   MN_Series ser;

   CHK_CUR;
   CHK_ARGS( !=, 4, "usage: grAgOp var list expr" )

   // Break the list up into elements, and execute the command once
   // for each value of the element.
   //

   result = Tcl_SplitList(interp, argv[2], &listArgc, &listArgv);
   if( result != TCL_OK ) return result;

   for( i = 0; i < listArgc; i++ )
   {
      if (Tcl_SetVar(interp, argv[1], listArgv[i], 0) == NULL)
      {
         Tcl_SetResult(interp, "couldn't set loop variable", TCL_STATIC);
         result = TCL_ERROR;
         break;
      }

      double d;
      result = Tcl_ExprDouble( interp, argv[3], &d );
      ser.Add( (float)d );
      if( result != TCL_OK ) break;
   }
   ckfree((char *) listArgv);
   if( result == TCL_OK )
   {
      sprintf( interp->result, "%g", ser.Stat( (MN_SeriesStat)data ) );
   }
   return result;
}

//------------------------------------------------------------------------------

COMMAND( grTC )
{
   CHK_CUR;
   CHK_ARGS( !=, 4, "usage: grTC v [forward|reverse|undirected] [A|B]" )

   Vertex *u, *v = (Vertex*)strtol( argv[1], (char**)NULL, 16 );
   if( DQ(type,v) != 0 ) RET_ERROR("object not vertex")
   char buf[16];
   int i, d;
   bool arc;

   if(      strcmp( argv[2], "forward"    ) == 0 ) d = 1;
   else if( strcmp( argv[2], "reverse"    ) == 0 ) d = 2;
   else if( strcmp( argv[2], "undirected" ) == 0 ) d = 3;
   else RET_ERROR("bad direction");
   if(      argv[3][0] == 'A' ) arc = true;
   else if( argv[3][0] == 'B' ) arc = false;
   else RET_ERROR("bad arc relation");

   VertexSet* vs = current_hi->hygraph->TC( v, d, arc );

   for( i = 0; (u = (*vs)[i]) != NIL; i++ )
   {
      sprintf( buf, PFMT(u) );
      Tcl_AppendElement( interp, buf );
   }

   delete vs;

   return TCL_OK;
}

//------------------------------------------------------------------------------

// Func: a(Obj,Index)	object attribute Index
// Func: na(Obj,Index)	object attribute Index
//
MATH_FUNC( grGetA )
{
   CHK_CUR;
   result->type = TCL_DOUBLE;
   HGObject* o = (HGObject*)args[0].intValue;
   int i = args[1].intValue;
   if( !InRange( i, 0, HG_MaxFIndx ) ) RET_ERROR("bad index");
   int normal = (int)data;
   if( normal )
   {
      int ty = DQ(type,o);
      if( current_hi->sdirty[ty] )
         current_hi->ComputeStats( ty, i );
      result->doubleValue = current_hi->stat[ty][i].Normal( DQ(val[i],o) );
   }
   else
      result->doubleValue = DQ(val[i],o);
   return TCL_OK;
}

//------------------------------------------------------------------------------

int GrRela_Init( Tcl_Interp* interp )
{
   Tcl_CreateCommand( interp, "grV",     grRel,    (ClientData)'V',
      (Tcl_CmdDeleteProc*)NULL);
   Tcl_CreateCommand( interp, "grA",     grRel,    (ClientData)'A',
      (Tcl_CmdDeleteProc*)NULL);
   Tcl_CreateCommand( interp, "grB",     grRel,    (ClientData)'B',
      (Tcl_CmdDeleteProc*)NULL);
   Tcl_CreateCommand( interp, "grF",     grRel,    (ClientData)'F',
      (Tcl_CmdDeleteProc*)NULL);
   Tcl_CreateCommand( interp, "grS",     grRel,    (ClientData)'S',
      (Tcl_CmdDeleteProc*)NULL);
   Tcl_CreateCommand( interp, "grE",     grRel,    (ClientData)'E',
      (Tcl_CmdDeleteProc*)NULL);

   Tcl_CreateCommand( interp, "grIn",    grSubRel, (ClientData)'I',
      (Tcl_CmdDeleteProc*)NULL);
   Tcl_CreateCommand( interp, "grOut",   grSubRel, (ClientData)'O',
      (Tcl_CmdDeleteProc*)NULL);
   Tcl_CreateCommand( interp, "grCB",    grSubRel, (ClientData)'B',
      (Tcl_CmdDeleteProc*)NULL);
   Tcl_CreateCommand( interp, "grCV",    grSubRel, (ClientData)'V',
      (Tcl_CmdDeleteProc*)NULL);

   Tcl_CreateCommand( interp, "grFrom",  grField,  (ClientData)'F',
      (Tcl_CmdDeleteProc*)NULL);
   Tcl_CreateCommand( interp, "grTo",    grField,  (ClientData)'T',
      (Tcl_CmdDeleteProc*)NULL);
   Tcl_CreateCommand( interp, "grId",    grField,  (ClientData)'I',
      (Tcl_CmdDeleteProc*)NULL);
   Tcl_CreateCommand( interp, "grLabel", grField,  (ClientData)'L',
      (Tcl_CmdDeleteProc*)NULL);
   Tcl_CreateCommand( interp, "grPB",    grField,  (ClientData)'B',
      (Tcl_CmdDeleteProc*)NULL);
   Tcl_CreateCommand( interp, "grPV",    grField,  (ClientData)'V',
      (Tcl_CmdDeleteProc*)NULL);

   Tcl_CreateCommand( interp, "grSelect", grSelect, (ClientData)NULL,
      (Tcl_CmdDeleteProc*)NULL);

   Tcl_CreateCommand( interp, "grTC",    grTC,     (ClientData)NULL,
      (Tcl_CmdDeleteProc*)NULL);

   Tcl_CreateCommand( interp, "grMin",   grAgOp,   (ClientData)StMin,
      (Tcl_CmdDeleteProc*)NULL);
   Tcl_CreateCommand( interp, "grMax",   grAgOp,   (ClientData)StMax,
      (Tcl_CmdDeleteProc*)NULL);
   Tcl_CreateCommand( interp, "grSum",   grAgOp,   (ClientData)StSum,
      (Tcl_CmdDeleteProc*)NULL);
   Tcl_CreateCommand( interp, "grAvg",   grAgOp,   (ClientData)StAverage,
      (Tcl_CmdDeleteProc*)NULL);
   Tcl_CreateCommand( interp, "grProd",  grAgOp,   (ClientData)StProduct,
      (Tcl_CmdDeleteProc*)NULL);

   Tcl_CreateCommand( interp, "grSet",   grSet,    (ClientData)NULL,
      (Tcl_CmdDeleteProc*)NULL);

   Tcl_ValueType a_ii[] = { TCL_INT, TCL_INT };

   Tcl_CreateMathFunc( interp, "a",  2, a_ii, grGetA, (ClientData)0 );
   Tcl_CreateMathFunc( interp, "na", 2, a_ii, grGetA, (ClientData)1 );

   return TCL_OK;
}

//------------------------------------------------------------------------------
